import functools
import pickle
import re

import numpy as np
from karabo.bound import (
    FLOAT_ELEMENT,
    IMAGEDATA_ELEMENT,
    INPUT_CHANNEL,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    UINT64_ELEMENT,
    ChannelMetaData,
    Dims,
    Encoding,
    Epochstamp,
    Hash,
    ImageData,
    MetricPrefix,
    Schema,
    Timestamp,
    Trainstamp,
    Unit,
)
from karabo.common.api import KARABO_SCHEMA_DISPLAY_TYPE_SCENES as DT_SCENES
from TrainMatcher import TrainMatcher, scenes as trainmatcher_scenes

from . import scenes
from ._version import version as deviceVersion
from calng import utils


preview_schema = Schema()
(
    IMAGEDATA_ELEMENT(preview_schema).key("image").commit(),
    UINT64_ELEMENT(preview_schema).key("trainId").readOnly().commit(),
)

xtdf_source_re = re.compile(r".*\/DET\/(\d+)CH0:xtdf")

# TODO: merge scene with TrainMatcher's nice overview
@KARABO_CLASSINFO("SimpleAssembler", deviceVersion)
class SimpleAssembler(TrainMatcher.TrainMatcher):
    @staticmethod
    def expectedParameters(expected):
        (
            OVERWRITE_ELEMENT(expected)
            .key("availableScenes")
            # TODO: add assemblerOverview scene
            .setNewDefaultValue(["scene", "assemblerOverview"])
            .commit(),
            FLOAT_ELEMENT(expected)
            .key("processingTime")
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .readOnly()
            .initialValue(0)
            .warnHigh(500)
            .info("Cannot keep up with GUI limit")
            .needsAcknowledging(False)
            .commit(),
            FLOAT_ELEMENT(expected)
            .key("timeOfFlight")
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .readOnly()
            .initialValue(0)
            .warnHigh(1000)
            .info("Time of flight exceeding 1 s")
            .needsAcknowledging(False)
            .commit(),
            STRING_ELEMENT(expected)
            .key("pathToStack")
            .assignmentOptional()
            .defaultValue("image.data")
            .commit(),
            INPUT_CHANNEL(expected)
            .key("geometryInput")
            .displayedName("Geometry input")
            .commit(),
            OUTPUT_CHANNEL(expected)  # can OVERWRITE_ELEMENT even do this?
            .key("output")
            .dataSchema(preview_schema)
            .commit(),
        )

    def initialization(self):
        super().initialization()

        # TODO: match inside device, fill multiple independent buffers

        self._path_to_stack = self.get("pathToStack")
        self.geometry = None
        self.input_buffer = None

        self.KARABO_ON_DATA("geometryInput", self.receive_geometry)
        self.KARABO_SLOT(self.requestScene)

        self.start()

    def requestScene(self, params):
        # TODO: unify with TrainMatcher overview
        scene_name = params.get("name", default="")
        if scene_name == "assemblerOverview":
            payload = Hash("name", scene_name, "success", True)
            payload["data"] = scenes.simple_assembler_overview(
                device_id=self.getInstanceId(),
                geometry_device_id=self.get("geometryInput.connectedOutputChannels")[
                    0
                ].split(":")[0],
            )
            self.reply(
                Hash(
                    "type",
                    "deviceScene",
                    "origin",
                    self.getInstanceId(),
                    "payload",
                    payload,
                )
            )
        else:
            return super().requestScene(params)

    def receive_geometry(self, data, metadata):
        self.log.INFO("Received a new geometry")
        self.geometry = pickle.loads(data.get("pickledGeometry"))
        self.input_buffer = np.zeros(self.geometry.expected_data_shape)

    def _send(self, train_id, sources):
        # TODO: adapt to appropriate hook for new TrainMatcher (no _send)
        if self.geometry is None:
            self.log.WARN("Have not received a geometry yet")
            return

        timestamp = Timestamp(Epochstamp(), Trainstamp(train_id))

        module_indices_unfilled = set(range(self.input_buffer.shape[0]))
        for source, (data, metadata) in sources.items():
            # TODO: handle failure to "parse" source, get data out
            module_index = self._source_to_index(source)
            self.input_buffer[module_index] = np.squeeze(data.get(self._path_to_stack))
            module_indices_unfilled.discard(module_index)

        for unfilled_module in module_indices_unfilled:
            self.input_buffer[module_index].fill(0)
            # TODO: configurable treatment of missing modules

        # TODO: reusable output buffer to save on allocation
        assembled, _ = self.geometry.position_modules_fast(self.input_buffer)

        # TODO: optionally include control data
        out_hash = Hash(
            "image",
            ImageData(
                # TODO: get around this being mirrored...
                (assembled[::-1, ::-1]).astype(np.int32),
                Dims(*assembled.shape),
                Encoding.GRAY,
            ),
            "trainId",
            train_id,
        )
        channel = self.signalSlotable.getOutputChannel("output")
        channel.write(out_hash, ChannelMetaData(self.getInstanceId(), timestamp))
        channel.update()
        self.rate_out.update()

    def _update_rate(self):
        self._buffered_status_update.set(
            "processingTime", self._processing_time_ema.get()
        )
        self._buffered_status_update.set("timeOfFlight", self._time_of_flight_ema.get())
        self._buffered_status_update.set("rate", self._rate_tracker.get())
        self.set(self._buffered_status_update)

    @functools.lru_cache
    def _source_to_index(self, source):
        # note: cache means warning only shows up once (also not performance-critical)
        match = xtdf_source_re.match(source)
        if match is None:
            self.log.WARN(f"Couldn't figure out index for source {source}")
            return 0
        return int(match.group(1))
