import pickle

from karabo.bound import (
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    IMAGEDATA_ELEMENT,
    NODE_ELEMENT,
    OUTPUT_CHANNEL,
    SLOT_ELEMENT,
    VECTOR_CHAR_ELEMENT,
    VECTOR_STRING_ELEMENT,
    Encoding,
    Hash,
    ImageData,
    PythonDevice,
    Schema,
    State,
)
from karabo.common.api import KARABO_SCHEMA_DISPLAY_TYPE_SCENES as DT_SCENES
import numpy as np
from matplotlib.backends.backend_agg import FigureCanvasAgg
import matplotlib.pyplot as plt

from . import scenes
from ._version import version as deviceVersion


geometry_schema = Schema()
(
    VECTOR_CHAR_ELEMENT(geometry_schema)
    .key("pickledGeometry")
    .displayedName("Pickled geometry")
    .assignmentOptional()
    .defaultValue([])
    .commit()
)

preview_schema = Schema()
(IMAGEDATA_ELEMENT(preview_schema).key("overview").commit())


@KARABO_CLASSINFO("ManualGeometryBase", deviceVersion)
class ManualGeometryBase(PythonDevice):
    @staticmethod
    def expectedParameters(expected):
        # "mandatory" for geometry serving device
        (
            OUTPUT_CHANNEL(expected)
            .key("geometryOutput")
            .dataSchema(geometry_schema)
            .commit(),
            SLOT_ELEMENT(expected).key("pleaseSendYourGeometry").commit(),
            SLOT_ELEMENT(expected).key("reloadScenes").commit(),
            OUTPUT_CHANNEL(expected)
            .key("previewOutput")
            .dataSchema(preview_schema)
            .commit(),
            IMAGEDATA_ELEMENT(expected).key("layoutPreview").commit(),
        )

        # configuring this one manually
        # subclasses should set better defaults
        (NODE_ELEMENT(expected).key("quadrantCorners").commit(),)
        for q in range(1, 5):
            (
                NODE_ELEMENT(expected).key(f"quadrantCorners.Q{q}").commit(),
                DOUBLE_ELEMENT(expected)
                .key(f"quadrantCorners.Q{q}.x")
                .assignmentOptional()
                .defaultValue(0)
                .reconfigurable()
                .commit(),
                DOUBLE_ELEMENT(expected)
                .key(f"quadrantCorners.Q{q}.y")
                .assignmentOptional()
                .defaultValue(0)
                .reconfigurable()
                .commit(),
            )

        # scenes are fun
        # TODO: add the scene
        (
            VECTOR_STRING_ELEMENT(expected)
            .key("availableScenes")
            .setSpecialDisplayType(DT_SCENES)
            .readOnly()
            .initialValue(["overview"])
            .commit(),
        )

    def __init__(self, config):
        super().__init__(config)

        self.KARABO_SLOT(self.pleaseSendYourGeometry)
        self.KARABO_SLOT(self.requestScene)
        self.KARABO_SLOT(self.reloadScenes)
        self.update_geom()
        plt.switch_backend("agg")
        self.updateState(State.ON)

    def requestScene(self, params):
        payload = Hash()
        scene_name = params.get("name", default="")
        payload["name"] = scene_name
        payload["success"] = True
        if scene_name == "overview":
            payload["data"] = scenes.manual_geometry_overview(
                device_id=self.getInstanceId()
            )
        else:
            payload["success"] = False
        response = Hash()
        response["type"] = "deviceScene"
        response["origin"] = self.getInstanceId()
        response["payload"] = payload
        self.reply(response)

    def update_geom(self):
        self.quadrant_corners = tuple(
            (self.get(f"quadrantCorners.Q{q}.x"), self.get(f"quadrantCorners.Q{q}.y"))
            for q in range(1, 5)
        )
        self.geom = self.geometry_class.from_quad_positions(self.quadrant_corners)
        self.pickled = pickle.dumps(self.geom)
        # TODO: send to anyone who asks? make slot for that?
        self.writeChannel("geometryOutput", Hash("pickledGeometry", self.pickled))

    def reloadScenes(self):
        global scenes
        import importlib

        scenes = importlib.reload(scenes)

    def pleaseSendYourGeometry(self):
        self.writeChannel("geometryOutput", Hash("pickledGeometry", self.pickled))
        axis = self.geom.inspect()
        axis.figure.tight_layout(pad=0)
        axis.figure.set_facecolor("none")
        # axis.figure.set_size_inches(6, 6)
        # axis.figure.set_dpi(300)
        canvas = FigureCanvasAgg(axis.figure)
        canvas.draw()
        image_buffer = np.frombuffer(canvas.buffer_rgba(), dtype=np.uint8).reshape(
            canvas.get_width_height()[::-1] + (4,)
        )
        self.set(
            "layoutPreview",
            ImageData(image_buffer, encoding=Encoding.RGBA, bitsPerPixel=3 * 8),
        )
        # self.writeChannel("previewOutput", Hash("overview", ImageData()))

    def preReconfigure(self, config):
        self._prereconfigure_update_hash = config

    def postReconfigure(self):
        if any(
            path.startswith("quadrantCorners")
            for path in self._prereconfigure_update_hash.getPaths()
        ):
            self.update_geom()
